/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.EakDefault
 *
 * de.bsvrz.kex.tls.osi7.conversion.EakDefault is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.EakDefault is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.EakDefault.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.EakDefault.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.EakDefault ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.EakDefault wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.osi7.conversion;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.kex.tls.TlsApplication;
import de.bsvrz.sys.funclib.debug.Debug;

//~ JDK IMPORTE ===============================================================

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

//~ KLASSEN ===================================================================

/**
 * Standard-Umsetzungsmodul für Einzeltelegramme am Eak
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class EakDefault implements EakConversionInterface {

    /**
     * UNDEF für Zustände, die nur nach dem Empfang eines entsprechenden Header-De-Blocks existieren.
     */
    static final int UNDEF = -1;

    /**
     * DebugLogger für Debug-Ausgaben
     */
    private static final Debug debug = Debug.getLogger();

    //~ FELDER ================================================================

    /**
     * Datenmodell der Konfiguration.
     */
    DataModel _config = null;

    // Objektvariablen für Zustände der FG 3 (UFD)
    // Objektvariablen für Zustände der FG 4 (WZG)
    private int _wzgLastSequenceNumber = UNDEF;
    private int _wzgNextSequenceNumber = UNDEF;

    // Objektvariablen für Zustände der FG 6 (VLT)
    // Objektvariablen für Zustände der FG 9 (ZFR)
    private int _zfrNextSequenceNumber = UNDEF;
    private int _zfrLastSequenceNumber = UNDEF;

    // Private ----------------------------------------------------------------
    private int _axlCurrentNumber;

    // Objektvariablen für Zustände der FG 2 (AXL)
    private int _axlIntervalDuration;
    private int _axlTotalNumber;

    /**
     * DaV Interface.
     */
    ClientDavInterface _daf;

    /**
     * Objektreferenz auf das aufrufende Eak-Objekt.
     */
    Eak _eak;

    /**
     * ID für das Einzeltelegramm.
     */
    private int _id;

    /**
     * Jobnummer für das Einzeltelegramm.
     */
    private int _jobNumber;

    /**
     * IntervallDauer der Ergebnisdaten der FG1 in Sekunden.
     */
    private long _lveIntervalDuration;

    // Objektvariablen für Zustände der FG 1 (LVE)

    /**
     * IntervallArt der Ergebnisdaten der FG1.
     */
    private int _lveIntervalType;

    /**
     * Objektreferenz auf das dem aufrufenden Eak-Objekt übergeordnete Sm-Objekt.
     */
    Sm _sm;

    // Objektvariablen für Zustände aller FG

    /**
     * Zeitstempel für das Einzeltelegramm aller Funktionsgruppen.
     */
    private long _timeStamp;

    /**
     * IntervallDauer der Ergebnisdaten der FG9 in Sekunden.
     */
    private long _zfrIntervalDuration;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    // Objektvariablen für Zustände der FG 254 (SYS)

    /**
     * Erzeugt eine neue Instanz von EakDefault
     * @param daf Datenverteiler
     * @param sm  Übergeordnetes SM
     * @param eak Das EAK
     */
    public EakDefault(ClientDavInterface daf, Sm sm, Eak eak) {
        _daf = daf;
        _sm  = sm;
        _eak = eak;

        // Nur zu testzwecken == null erlaubt!
        if (_daf != null) {
            _config = _daf.getDataModel();
        }

        debug.finer("EakDefault Klasse wurde erfolgreich geladen und instanziert.");
    }

    //~ GET METHODEN ==========================================================

    // ------------------------------------------------------------------------
    // GETTER / SETTER METHODEN
    // ------------------------------------------------------------------------

    /**
     * Liefert die laufende Nummer des Datensatzes der Einzelergebnisdaten der FG2. Entspricht dem Wert <code>laufende
     * Nummer des Datensatzes</code> im Typ 31 der FG 2.
     *
     * @return Laufende Nummer des Datensatzes.
     */
    public int getAxlCurrentNumber() {
        return _axlCurrentNumber;
    }

    /**
     * Liefert die IntervallDauer der Ergebnisdaten der FG2 in Sekunden. Entspricht dem Wert <code>Intervalllänge</code>
     * im Typ 64 der FG 2.
     *
     * @return IntervallDauer in Sekungen.
     */
    public int getAxlIntervalDuration() {
        return _axlIntervalDuration;
    }

    /**
     * Liefert die Gesamtzahl Datensatzes der Einzelergebnisdaten der FG2. Entspricht dem Wert <code>Gesamtzahl der
     * Datensätze</code> im Typ 31 der FG 2.
     *
     * @return Gesamtzahl der Datensätze.
     */
    public int getAxlTotalNumber() {
        return _axlTotalNumber;
    }

    /**
     * Ermöglicht den Zugriff auf die aktuelle Datenverteilerverbindung.
     *
     * @return Gibt das aktuelle aktuelle Datenverteilerverbindung zurück.
     */
    public ClientDavInterface getClientDavInterface() {
        return _daf;
    }

    /**
     * Liefert die Jobnummer des Einzeltelegramms.
     *
     * @return Jobnummer für das Einzeltelegramm (0=spontan, 1..255=Antwort auf Telegramm mit dieser Jobnummer).
     */
    public int getJobNumber() {
        return _jobNumber;
    }

    /**
     * Liefert die IntervallDauer der Ergebnisdaten der FG1 in Sekunden. Entspricht dem Wert <code>Intervalllänge</code>
     * im Typ 48 und 64 der FG 1.
     *
     * @return IntervallDauer in Sekungen.
     */
    public long getLveIntervalDuration() {
        return _lveIntervalDuration;
    }

    /**
     * Liefert die IntervallArt der Ergebnisdaten der FG1. Entspricht dem Wert <code>Art des Intervalls</code> im Typ 48
     * und 64 der FG 1.
     *
     * @return Intervalltyp: 1 = Kurzzeitdaten, 3 = streckenbezogene Daten
     */
    public int getLveIntervalType() {
        return _lveIntervalType;
    }

    /**
     * Liefert den Zeitstempel für das Einzeltelegramm aller Funktionsgruppen. Der Wert ist die Zeitdifferenz zum
     * 01.01.1970 (Mitternacht) in Millisekunden in der Zeitzone UTC.
     *
     * @return Zeitstempel für das Einzeltelegramm.
     */
    public long getTimeStamp() {
        return _timeStamp;
    }

    /**
     * Liefert die aktuelle Folgenummer der FG4. Entspricht dem Wert <code>Folgenummer</code> im Typ 31 der FG 4.
     *
     * @return Folgenummer der Datensätze.
     */
    public int getWzgNextSequenceNumber() {
        return _wzgNextSequenceNumber;
    }

    /**
     * Liefert die IntervallDauer der Ergebnisdaten der FG9 in Sekunden. Entspricht dem Wert <code>Intervalllänge</code>
     * im Typ 64 der FG 9.
     *
     * @return IntervallDauer in Sekungen.
     */
    public long getZfrIntervalDuration() {
        return _zfrIntervalDuration;
    }

    /**
     * Liefert die aktuelle Folgenummer der FG9. Entspricht dem Wert <code>Folgenummer</code> im Typ 31 der FG 9.
     *
     * @return Folgenummer der Datensätze.
     */
    public int getZfrNextSequenceNumber() {
        return _zfrNextSequenceNumber;
    }

    //~ METHODEN ==============================================================

    /**
     * Verarbeitet ein EinzelTelegramm. Enthält das Einzeltelegramm als ersten De-Block einen De-Block für die
     * DeKanalNummer 255, so werden die Informationen dieses Blocks hier durch das Eak ausgewertet (Zeitinformationen
     * und Intervalldaten etc.). Die für das Einzeltelegramm ermittelten allgemeinen Information stehen dann in den
     * Umsetzungsmodulen für die De allen De's des Einzeltelegramms zur Verfügung.
     *
     * @param singleTelegram Das zu verarbeitende Einzeltelegramm {@link SingleTelegram}.
     */
    public void processSingleTelegram(SingleTelegram singleTelegram) {

        // Jobnummer auslesen
        _jobNumber = singleTelegram.getJobNumber();
        _id        = singleTelegram.getId();

        // Objektvariablen für Zustände aller FG initialisieren.
        setTimeStamp(System.currentTimeMillis());  // auf aktuelle Uhrzeit setzen
        debug.finest("Setze Zeitstempel auf: " + new Date(getTimeStamp()).toString());

        // Objektvariablen für Zustände der FG 1 (LVE) initialisieren.
        _lveIntervalType     = UNDEF;
        _lveIntervalDuration = UNDEF;

        // Objektvariablen für Zustände der FG 2 (AXL) initialisieren.
        _axlIntervalDuration = UNDEF;
        _axlCurrentNumber    = UNDEF;
        _axlTotalNumber      = UNDEF;

        // Objektvariablen für Zustände der FG 3 (UFD) initialisieren.
        // Objektvariablen für Zustände der FG 4 (WZG) initialisieren.
        // Objektvariablen für Zustände der FG 6 (VLT) initialisieren.
        // Objektvariablen für Zustände der FG 9 (ZFR) initialisieren.
        _zfrIntervalDuration = UNDEF;

        // Objektvariablen für Zustände der FG 254 (SYS) initialisieren.
        // Wenn im ersten DE-Block De=255, dann erfolgt Verarbeitung durch den EAK, sonst durch den DE-Block
        // Neu seit TLS 2012: Wenn erster DE-Block FG2 und Typ31 erfolgt auch hier die Verarbeitung, m. E.
        // ein Fehler in der TLS, aber so wurde sie offiziell veröffentlicht.
        if (((singleTelegram.getCountDeBlock() > 0) && (singleTelegram.getDeBlock()[0].getDeChannel() == 255))
            ||

        // NEU wegen Fehler in TLS 2012
        (singleTelegram.getCountDeBlock() > 0) && (singleTelegram.getDeBlock()[0].getFg() == 2) && (singleTelegram.getDeBlock()[0].getTyp() == 31)) {
            DeBlock deBlock = singleTelegram.getDeBlock()[0];

            debug.finest("Verabeitung durch den EAK: " + Debug.NEWLINE + deBlock.toString());

            switch (deBlock.getFg()) {
            case 1 :
                switch (deBlock.getTyp()) {
                case 30 :
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;

                case 48 :                                                                // Intervalldaten Kurz- und streckenbezogene Daten
                    processTimeStampHMS(deBlock.getDeData().readNextTlsDataBlock(3));    // Zeitstempel
                    _lveIntervalType     = deBlock.getDeData().readUnsignedByteAsInt();
                    _lveIntervalDuration = deBlock.getDeData().readUnsignedByteAsInt() * 15;

                    break;

                case 64 :                                                                // Intervalldaten Langzeit
                    processTimeStampYMDH(deBlock.getDeData().readNextTlsDataBlock(4));   // Zeitstempel
                    _lveIntervalDuration = ((deBlock.getDeData().readUnsignedByteAsInt() & 0x7f) * 3600);

                    break;
                }

                break;

            case 2 :
                switch (deBlock.getTyp()) {
                case 30 :
                    debug.finest("Hier darf ich nicht vorbei kommen: Typ 30 Zeitstempel in der FG 2 !");
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;

                case 31 :                                                                // Zeitstempel für Einzelergebnismeldungen
                    processTimeStampYYMDHMShs(deBlock.getDeData().readNextTlsDataBlock(8));
                    _axlCurrentNumber = deBlock.getDeData().readUnsignedLowHighByteAsInt();
                    _axlTotalNumber   = deBlock.getDeData().readUnsignedLowHighByteAsInt();

                    break;

                case 64 :                                                                // Intervalldaten Achslast
                    processTimeStampYMDHM(deBlock.getDeData().readNextTlsDataBlock(5));  // Zeitstempel
                    _axlIntervalDuration = deBlock.getDeData().readUnsignedByteAsInt();

                    if ((_axlIntervalDuration & 0x80) != 0) {                            // Stundenbit gesetzt
                        _axlIntervalDuration = (_axlIntervalDuration & 0x7f) * 3600;
                    } else {                                                             // Wert in 1/4 Stunden Einheit
                        _axlIntervalDuration = (_axlIntervalDuration & 0x7f) * 3600 / 4;
                    }

                    break;
                }

                break;

            case 3 :
                switch (deBlock.getTyp()) {
                case 30 :
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;
                }

                break;

            case 4 :
                switch (deBlock.getTyp()) {
                case 30 :
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;

                case 31 :                                                                // Zeitstempel mit Folgenummer
                    processTimeStampHMSDWithSequenceNumberFg4(deBlock);

                    break;
                }

                break;

            case 6 :
                switch (deBlock.getTyp()) {
                case 30 :
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;
                }

                break;

            case 9 :
                switch (deBlock.getTyp()) {
                case 30 :
                    debug.finest("Hier darf ich nicht vorbei kommen: Typ 30 Zeitstempel in der FG 9 !");
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;

                case 31 :                                                                // Zeitstempel mit Folgenummer
                    processTimeStampHMSDWithSequenceNumberFg9(deBlock);

                    break;

                case 64 :                                                                // Intervalldaten ZFR
                    processTimeStampHMS(deBlock.getDeData().readNextTlsDataBlock(3));    // Zeitstempel
                    _zfrIntervalDuration = deBlock.getDeData().readUnsignedByteAsInt() * 15;

                    break;
                }

                break;

            case 254 :
                switch (deBlock.getTyp()) {
                case 30 :
                    processTimeStampHMS(deBlock.getDeData());                            // Zeitstempel

                    break;
                }

                break;

            default :
                debug.finest(String.format("Keine Auswertung für Header-De-Block (De=255) in Fg %d bei Typ %d", deBlock.getFg(), deBlock.getTyp()));
            }
        }
    }

    // ------------------------------------------------------------------------
    // SONSTIGE METHODEN
    // ------------------------------------------------------------------------

    /**
     * Wertet den Zeitanteil eines DE-Blocks bestehend aus 3 Byte Stunde/Minute/Sekunde aus und stellt das Ergebnis über
     * die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
     */
    void processTimeStampHMS(TlsData _tlsData) {

        // Byte aus Telegramm -------------------------------------------------------------------------------
        int _ZeitbereichStunde_ = _tlsData.readUnsignedByteAsInt();
        int _Minute_            = _tlsData.readUnsignedByteAsInt();
        int _Sekunde_           = _tlsData.readUnsignedByteAsInt();

        setTimeStamp(-1, -1, -1, _ZeitbereichStunde_, _Minute_, _Sekunde_, 0);
    }

    /**
     * Wertet den Zeitanteil eines DE-Blocks bestehend aus 4 Byte Stunde/Minute/Sekunde/Tag aus und stellt das Ergebnis
     * über die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
     */
    void processTimeStampHMSD(TlsData _tlsData) {

        // Byte aus Telegramm -------------------------------------------------------------------------------
        int _ZeitbereichStunde_ = _tlsData.readUnsignedByteAsInt();
        int _Minute_            = _tlsData.readUnsignedByteAsInt();
        int _Sekunde_           = _tlsData.readUnsignedByteAsInt();
        int _Tag_               = _tlsData.readUnsignedByteAsInt();

        setTimeStamp(-1, -1, _Tag_, _ZeitbereichStunde_, _Minute_, _Sekunde_, 0);
    }

    /**
     * Verarbeite die Zeitstempel mit Folgenummern der FG 4 /Typ 31. <ol> <li>Für alle potentiell betroffenen ATG leeren
     * Datensatz versenden. Dazu Liste ermitteln, indem alle angemeldeten Attributgruppen für die relevanten FG 4 DE und
     * die darauf referenzierenden Objekte an diesem Eak ermittelt werden. </li> <li>Abfrage aller aktuellen Stati aller
     * potentiell fehlenden Telegramme, da durch senden des leeren Datensatzes kein akteuller Zustand mehr existiert.
     * </li> <li>Fehlende Nummern nachfordern und als "nachgeliefert" abspeichern.</li> </ol>
     *
     * @param deBlock Der Zeitstempel mit Folgenummern De-Block der FG 4 / Typ 31.
     */
    void processTimeStampHMSDWithSequenceNumberFg4(DeBlock deBlock) {
        processTimeStampHMSD(deBlock.getDeData().readNextTlsDataBlock(4));  // Zeitstempel

        if (TlsApplication.isControlLostTelegramsFg4()) {

            // Folgenummer NICHT auswerten, wenn 32 < ID < 48 ist, da Folgennummer
            // aus Pufferabruf stammt und nicht berücksichtigt werden darf.
            if (!((_id > 32)
                  && (_id < 48))) {
                _wzgLastSequenceNumber = _wzgNextSequenceNumber;                              // letzte Folgenummer merken

                if (_wzgLastSequenceNumber == 65535) {
                    _wzgLastSequenceNumber = 0;                                               // da nächste Nummer = 1 (0 ist nicht zulässig!!)
                }

                _wzgNextSequenceNumber = deBlock.getDeData().readUnsignedLowHighByteAsInt();  // Aktuelle Folgenummer

                if ((_wzgLastSequenceNumber != UNDEF)
                    && (_wzgNextSequenceNumber - _wzgLastSequenceNumber) != 1) {
                    int wzgMissingNumber = _wzgNextSequenceNumber - _wzgLastSequenceNumber - 1;

                    // Überlauf testen
                    if (wzgMissingNumber < 0) {
                        wzgMissingNumber += 65535 - 1;
                    }

                    debug.warning("Folgenummernfehler. Empfangene Nummer = " + _wzgNextSequenceNumber + ", letze Nummer = " + _wzgLastSequenceNumber + ", d.h. es fehlen " + (wzgMissingNumber) + " Telegramme an (Osi7Adresse/De/Fg):" + Debug.NEWLINE + deBlock.getDescriptionKey());

                    // Für alle potentiell betroffenen ATG leeren Datensatz versenden. Liste ermitteln, indem alle
                    // angemeldeten Attributgruppen für die relevanten FG 4 DE und die darauf referenzierenden Objekte
                    // an diesem Eak ermittelt werden.
                    long timeStamp = System.currentTimeMillis();  // aktuelle Zeit als Zeitstempel verwenden
                    List deList    = _eak.getDeList();            // Liste aller De  an diesem EAK holen

                    // Prüfen, ob es eine DE255 der FG4 an diesem EAK gibt, dann erfolgt die Abfrage später für diese DE,
                    // sonst für alle DE ausser CLustern einzeln
                    boolean de255Fg4Exists = false;

                    for (int i = 0; i < deList.size(); i++) {
                        De de = (De) deList.get(i);

                        if (de.getDeChannel() == 255) {
                            de255Fg4Exists = true;

                            break;
                        }
                    }

                    for (int i = 0; i < deList.size(); i++) {
                        De de = (De) deList.get(i);

                        // Für alle De der FG 4 an diesem EAK....
                        if (de.getFg() == 4) {
                            DeConversionModul deCM         = de.getConversionModul();
                            List              tlsToDavList = deCM.getTlsToDavList();

                            // Für alle Einträge an diesem De der FG 4, ...
                            for (int j = 0; j < tlsToDavList.size(); j++) {
                                Object[]        tlsToDav        = (Object[]) tlsToDavList.get(j);
                                AttributeGroup  attributeGroup  = _config.getAttributeGroup((String) tlsToDav[DeConversionModul.ATG]);
                                Aspect          aspect          = _config.getAspect((String) tlsToDav[DeConversionModul.ASP]);
                                DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
                                Data            data            = null;  // Leeren Datensatz erzeugen

                                // Jetzt für alle angemeldeten Objekte auf diese Konvertierung einen ResultData
                                // erzeugen und versenden (Austausch des Zielobjekts):
                                SystemObject[] objects = de.getObjectRefsToDe(_config.getType((String) tlsToDav[DeConversionModul.OT]));
                                ResultData[]   results = new ResultData[objects.length];

                                // Leeren Datensatz senden...
                                for (int k = 0; k < objects.length; ++k) {
                                    results[k] = new ResultData(objects[k], dataDescription, timeStamp, data);
                                    debug.fine("Sende leeren Datensatz wegen Folgenummernfehler an DaV: " + objects[k].toString() + dataDescription.toString());
                                    deCM.sendData(results);
                                }
                            }

                            // Aktuellen Status direkt für alle abfragen die zuvor auf "keine Daten" gesetzt wurden.
                            AttributeGroup  attributeGroup  = _config.getAttributeGroup("atg.tlsGloDeBlockAbruf");
                            Aspect          aspect          = _config.getAspect("asp.tlsAbruf");
                            DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
                            Data            data            = _daf.createData(attributeGroup);

                            // Typ/ID Kombinationen der FG 4, die bei Folgenummernfehler an DE255 der FG4 abgefragt werden müssen
                            int[][] missingData255 = new int[][] {
                                { 1, 1 },   // De-Fehler
                                { 2, 1 },   // Nicht darstellbare WVZ
                                { 4, 1 },   // Gestörte Textpositionen
                                { 14, 1 },  // Ergänzende DE-Fehler
                                { 17, 2 },  // Betriebsart
                                { 29, 2 },  // Kanal-Steuerung
                                { 33, 3 },  // Grundeinstellung
                                { 44, 3 },  // Code Definition
                                { 48, 5 },  // WVZ Stellzustand
                                { 49, 5 },  // Helligkeit
                                { 50, 5 },  // Wechseltext
                                { 55, 5 },  // Stellzustand
                                { 58, 5 },  // Stellprogramm
                            };

                            // Typ/ID Kombinationen der FG 4, die bei Folgenummernfehler an DE ungleich DE255 und kein Cluster abgefragt werden müssen
                            int[][] missingDataDe = new int[][] {
                                { 1, 1 },   // De-Fehler
                                { 2, 1 },   // Nicht darstellbare WVZ
                                { 4, 1 },   // Gestörte Textpositionen
                                { 14, 1 },  // Ergänzende DE-Fehler
                                { 29, 2 },  // Kanal-Steuerung
                                { 33, 3 },  // Grundeinstellung
                                { 44, 3 },  // Code Definition
                                { 48, 5 },  // WVZ Stellzustand
                                { 50, 5 },  // Wechseltext
                                { 55, 5 },  // Stellzustand
                                { 58, 5 },  // Stellprogramm
                            };

                            // Typ/ID Kombinationen der FG 4, die bei Folgenummernfehler an Clusterkanälen abgefragt werden müssen
                            int[][] missingDataCluster = new int[][] {
                                { 1, 1 },                                                            // De-Fehler
                                { 14, 1 },                                                           // Ergänzende DE-Fehler
                                { 17, 2 },                                                           // Betriebsart
                                { 29, 2 },                                                           // Kanal-Steuerung
                                { 33, 3 },                                                           // Grundeinstellung
                                { 44, 3 },                                                           // Code Definition
                                { 49, 5 },                                                           // Helligkeit
                                { 55, 5 },                                                           // Stellzustand
                                { 58, 5 },                                                           // Stellprogramm
                            };

                            // Abfrage der fehlenden Daten nur an der DE255, sofern diese existiert
                            // sonst an jedem DE, welches kein Cluster ist abfragen
                            if (de255Fg4Exists
                                && (de.getDeChannel() == 255)) {
                                for (int j = 0; j < missingData255.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingData255[j][0]);          // Typ
                                    data.getUnscaledValue("ID").set(missingData255[j][1] + 16);      // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // Wenn keine DE255 existiert UND De ist kein Cluster, dann Abfrage für DE
                            if (!de255Fg4Exists
                                &&!de.isCluster()) {
                                for (int j = 0; j < missingDataDe.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingDataDe[j][0]);           // Typ
                                    data.getUnscaledValue("ID").set(missingDataDe[j][1] + 16);       // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // Wenn keine DE255 existiert UND De ist  Cluster, dann Abfrage für Cluster
                            if (!de255Fg4Exists
                                && de.isCluster()) {
                                for (int j = 0; j < missingDataCluster.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingDataCluster[j][0]);      // Typ
                                    data.getUnscaledValue("ID").set(missingDataCluster[j][1] + 16);  // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // ... und zum Schluss noch die fehlenden Folgenummern anfordern...
                            if (de.isCluster()) {
                                debug.warning("De, an dem Folgenummern abgefragt werden: " + de.toString());
                                attributeGroup  = _config.getAttributeGroup("atg.tlsWzgAbrufPufferInhalt");
                                aspect          = _config.getAspect("asp.tlsAbruf");
                                dataDescription = new DataDescription(attributeGroup, aspect);
                                data            = _daf.createData(attributeGroup);
                                data.getUnscaledValue("StartFolgeNummer").set(_wzgLastSequenceNumber + 1);

                                // Es werden maximal 255 Telegramme nachgefordert, auch bei echtem Rücksprung (statt der 65xxx Tausend).
                                data.getUnscaledValue("Anzahl1").set((wzgMissingNumber > 255) ? 255 : wzgMissingNumber);

                                ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                debug.warning("Resultdata für Folgennummernabfrage: " + result);
                                deCM.sendData(result);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Verarbeite die Zeitstempel mit Folgenummern der FG 9 /Typ 31. <ol> <li>Für alle potentiell betroffenen ATG leeren
     * Datensatz versenden. Dazu Liste ermitteln, indem alle angemeldeten Attributgruppen für die relevanten FG 9 DE und
     * die darauf referenzierenden Objekte an diesem Eak ermittelt werden. </li> <li>Abfrage aller aktuellen Stati aller
     * potentiell fehlenden Telegramme, da durch senden des leeren Datensatzes kein akteuller Zustand mehr existiert.
     * </li> <li>Fehlende Nummern nachfordern und als "nachgeliefert" abspeichern.</li> </ol>
     *
     * @param deBlock Der Zeitstempel mit Folgenummern De-Block der FG 9 / Typ 31.
     */
    void processTimeStampHMSDWithSequenceNumberFg9(DeBlock deBlock) {
        processTimeStampHMSD(deBlock.getDeData().readNextTlsDataBlock(4));  // Zeitstempel

        if (TlsApplication.isControlLostTelegramsFg9()) {

            // Folgenummer NICHT auswerten, wenn 32 < ID < 48 ist, da Folgennummer
            // aus Pufferabruf stammt und nicht berücksichtigt werden darf.
            if (!((_id > 32)
                  && (_id < 48))) {
                _zfrLastSequenceNumber = _zfrNextSequenceNumber;                              // letzte Folgenummer merken

                if (_zfrLastSequenceNumber == 65535) {
                    _zfrLastSequenceNumber = 0;                                               // da nächste Nummer = 1 (0 ist nicht zulässig!!)
                }

                _zfrNextSequenceNumber = deBlock.getDeData().readUnsignedLowHighByteAsInt();  // Aktuelle Folgenummer

                if ((_zfrLastSequenceNumber != UNDEF)
                    && (_zfrNextSequenceNumber - _zfrLastSequenceNumber) != 1) {
                    int zfrMissingNumber = _zfrNextSequenceNumber - _zfrLastSequenceNumber - 1;

                    // Überlauf testen
                    if (zfrMissingNumber < 0) {
                        zfrMissingNumber += 65535 - 1;
                    }

                    debug.warning("Folgenummernfehler. Empfangene Nummer = " + _zfrNextSequenceNumber + ", letze Nummer = " + _zfrLastSequenceNumber + ", d.h. es fehlen " + (zfrMissingNumber) + " Telegramme an (Osi7Adresse/De/Fg):" + Debug.NEWLINE + deBlock.getDescriptionKey());

                    // Für alle potentiell betroffenen ATG leeren Datensatz versenden. Liste ermitteln, indem alle
                    // angemeldeten Attributgruppen für die relevanten FG 9 DE und die darauf referenzierenden Objekte
                    // an diesem Eak ermittelt werden.
                    long timeStamp = System.currentTimeMillis();  // aktuelle Zeit als Zeitstempel verwenden
                    List deList    = _eak.getDeList();            // Liste aller De  an diesem EAK holen

                    // Prüfen, ob es eine DE255 der FG9 an diesem EAK gibt, dann erfolgt die Abfrage später für diese DE,
                    // sonst für alle DE ausser Clustern einzeln
                    boolean de255Fg9Exists = false;

                    for (int i = 0; i < deList.size(); i++) {
                        De de = (De) deList.get(i);

                        if (de.getDeChannel() == 255) {
                            de255Fg9Exists = true;

                            break;
                        }
                    }

                    for (int i = 0; i < deList.size(); i++) {
                        De de = (De) deList.get(i);

                        // Für alle De der FG 9 an diesem EAK....
                        if (de.getFg() == 9) {
                            DeConversionModul deCM         = de.getConversionModul();
                            List              tlsToDavList = deCM.getTlsToDavList();

                            // Für alle Einträge an diesem De der FG 9, ...
                            for (int j = 0; j < tlsToDavList.size(); j++) {
                                Object[]        tlsToDav        = (Object[]) tlsToDavList.get(j);
                                AttributeGroup  attributeGroup  = _config.getAttributeGroup((String) tlsToDav[DeConversionModul.ATG]);
                                Aspect          aspect          = _config.getAspect((String) tlsToDav[DeConversionModul.ASP]);
                                DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
                                Data            data            = null;  // Leeren Datensatz erzeugen

                                // Jetzt für alle angemeldeten Objekte auf diese Konvertierung einen ResultData
                                // erzeugen und versenden (Austausch des Zielobjekts):
                                SystemObject[] objects = de.getObjectRefsToDe(_config.getType((String) tlsToDav[DeConversionModul.OT]));
                                ResultData[]   results = new ResultData[objects.length];

                                // Leeren Datensatz senden...
                                for (int k = 0; k < objects.length; ++k) {
                                    results[k] = new ResultData(objects[k], dataDescription, timeStamp, data);
                                    debug.fine("Sende leeren Datensatz wegen Folgenummernfehler an DaV: " + objects[k].toString() + dataDescription.toString());
                                    deCM.sendData(results);
                                }
                            }

                            // Aktuellen Status direkt für alle abfragen die zuvor auf "keine Daten" gesetzt wurden.
                            AttributeGroup  attributeGroup  = _config.getAttributeGroup("atg.tlsGloDeBlockAbruf");
                            Aspect          aspect          = _config.getAspect("asp.tlsAbruf");
                            DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
                            Data            data            = _daf.createData(attributeGroup);

                            // Typ/ID Kombinationen der FG 9, die bei Folgenummernfehler an DE255 der FG9 abgefragt werden müssen
                            int[][] missingData255 = new int[][] {
                                { 1, 1 },   // De-Fehler
                                { 14, 1 },  // Ergänzende DE-Fehler
                                { 17, 2 },  // Betriebsart
                                { 29, 2 },  // Kanal-Steuerung
                                { 32, 3 },  // Betriebsparameter Zuflussregelung
                                { 33, 3 },  // Betriebsparameter ALINEA1
                                { 34, 3 },  // Betriebsparameter Rotfahrerzähler
                                { 35, 3 },  // Betriebsparameter ALINEA2
                                { 48, 5 },  // Signalplan
                                { 49, 5 },  // Helligkeit
                                { 65, 4 },  // Rotfahrerzähler
                                { 70, 4 },  // ALINEA Schaltvorschläge
                            };

                            // Typ/ID Kombinationen der FG 9, die bei Folgenummernfehler an DE ungleich DE255 und kein Cluster abgefragt werden müssen
                            int[][] missingDataDe = new int[][] {
                                { 1, 1 },   // De-Fehler
                                { 14, 1 },  // Ergänzende DE-Fehler
                                { 29, 2 },  // Kanal-Steuerung
                                { 32, 3 },  // Betriebsparameter Zuflussregelung
                                { 33, 3 },  // Betriebsparameter ALINEA1
                                { 34, 3 },  // Betriebsparameter Rotfahrerzähler
                                { 35, 3 },  // Betriebsparameter ALINEA2
                                { 48, 5 },  // Signalplan
                                { 65, 4 },  // Rotfahrerzähler
                                { 70, 4 },  // ALINEA Schaltvorschläge
                            };

                            // Typ/ID Kombinationen der FG 9, die bei Folgenummernfehler an Clusterkanälen abgefragt werden müssen
                            int[][] missingDataCluster = new int[][] {
                                { 1, 1 },                                                            // De-Fehler
                                { 14, 1 },                                                           // Ergänzende DE-Fehler
                                { 17, 2 },                                                           // Betriebsart
                                { 29, 2 },                                                           // Kanal-Steuerung
                                { 32, 3 },                                                           // Betriebsparameter Zuflussregelung
                                { 33, 3 },                                                           // Betriebsparameter ALINEA1
                                { 34, 3 },                                                           // Betriebsparameter Rotfahrerzähler
                                { 35, 3 },                                                           // Betriebsparameter ALINEA2
                                { 48, 5 },                                                           // Signalplan
                                { 49, 5 },                                                           // Helligkeit
                                { 65, 4 },                                                           // Rotfahrerzähler
                                { 70, 4 },                                                           // ALINEA Schaltvorschläge
                            };

                            // Abfrage der fehlenden Daten nur an der DE255, sofern diese existiert
                            // sonst an jedem DE, welches kein Cluster ist abfragen
                            if (de255Fg9Exists
                                && (de.getDeChannel() == 255)) {
                                for (int j = 0; j < missingData255.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingData255[j][0]);          // Typ
                                    data.getUnscaledValue("ID").set(missingData255[j][1] + 16);      // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // Wenn keine DE255 existiert UND De ist kein Cluster, dann Abfrage für DE
                            if (!de255Fg9Exists
                                &&!de.isCluster()) {
                                for (int j = 0; j < missingDataDe.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingDataDe[j][0]);           // Typ
                                    data.getUnscaledValue("ID").set(missingDataDe[j][1] + 16);       // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // Wenn keine DE255 existiert UND De ist  Cluster, dann Abfrage für Cluster
                            if (!de255Fg9Exists
                                && de.isCluster()) {
                                for (int j = 0; j < missingDataCluster.length; j++) {
                                    data.getUnscaledValue("JobNummer").set(27);                      // wird nicht ausgewertet, steht aber im DatKat. Jobnummer wird beim versenden automatisch erzeugt.
                                    data.getUnscaledValue("Typ").set(missingDataCluster[j][0]);      // Typ
                                    data.getUnscaledValue("ID").set(missingDataCluster[j][1] + 16);  // Bei Abruf ID + 16

                                    ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                    deCM.sendData(result);
                                }
                            }

                            // ... und zum Schluss noch die fehlenden Folgenummern anfordern...
                            if (de.isCluster()) {
                                debug.warning("De, an dem Folgenummern abgefragt werden: " + de.toString());
                                attributeGroup  = _config.getAttributeGroup("atg.tlsZfrAbrufPufferInhalt");
                                aspect          = _config.getAspect("asp.tlsAbruf");
                                dataDescription = new DataDescription(attributeGroup, aspect);
                                data            = _daf.createData(attributeGroup);
                                data.getUnscaledValue("StartFolgeNummer").set(_zfrLastSequenceNumber + 1);

                                // Es werden maximal 255 Telegramme nachgefordert, auch bei echtem Rücksprung (statt der 65xxx Tausend).
                                data.getUnscaledValue("Anzahl1").set((zfrMissingNumber > 255) ? 255 : zfrMissingNumber);

                                ResultData result = new ResultData(de.getDe(), dataDescription, timeStamp, data);

                                debug.warning("Resultdata für Folgennummernabfrage: " + result);
                                deCM.sendData(result);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Wertet den Zeitanteil eines DE-Blocks bestehend aus 4 Byte Jahr/Monat/Tag/Stunde aus und stellt das Ergebnis über
     * die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
     */
    void processTimeStampYMDH(TlsData _tlsData) {

        // Byte aus Telegramm -------------------------------------------------------------------------------
        int _Jahr_              = _tlsData.readUnsignedByteAsInt();
        int _Monat_             = _tlsData.readUnsignedByteAsInt();
        int _Tag_               = _tlsData.readUnsignedByteAsInt();
        int _ZeitbereichStunde_ = _tlsData.readUnsignedByteAsInt();

        setTimeStamp(_Jahr_, _Monat_, _Tag_, _ZeitbereichStunde_, 0, 0, 0);
    }

    /**
     * Wertet den Zeitanteil eines DE-Blocks bestehend aus 5 Byte Jahr/Monat/Tag/Stunde/Minute aus und stellt das
     * Ergebnis über die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
     */
    void processTimeStampYMDHM(TlsData _tlsData) {

        // Byte aus Telegramm -------------------------------------------------------------------------------
        int _Jahr_              = _tlsData.readUnsignedByteAsInt();
        int _Monat_             = _tlsData.readUnsignedByteAsInt();
        int _Tag_               = _tlsData.readUnsignedByteAsInt();
        int _ZeitbereichStunde_ = _tlsData.readUnsignedByteAsInt();
        int _Minute_            = _tlsData.readUnsignedByteAsInt();

        setTimeStamp(_Jahr_, _Monat_, _Tag_, _ZeitbereichStunde_, _Minute_, 0, 0);
    }

    /**
     * Wertet den Zeitanteil eines DE-Blocks bestehend aus 8 Byte Jahr(2 Byte)/Monat/Tag/Stunde/Miute/Sekunde/HunderstelSekund
     * aus und stellt das Ergebnis über die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
     */
    void processTimeStampYYMDHMShs(TlsData _tlsData) {

        // Byte aus Telegramm -------------------------------------------------------------------------------
        int _Jahr_              = _tlsData.readUnsignedLowHighByteAsInt();
        int _Monat_             = _tlsData.readUnsignedByteAsInt();
        int _Tag_               = _tlsData.readUnsignedByteAsInt();
        int _ZeitbereichStunde_ = _tlsData.readUnsignedByteAsInt();
        int _Minute_            = _tlsData.readUnsignedByteAsInt();
        int _Sekunde_           = _tlsData.readUnsignedByteAsInt();
        int _HunderstelSekunde_ = _tlsData.readUnsignedByteAsInt();

        setTimeStamp(_Jahr_, _Monat_, _Tag_, _ZeitbereichStunde_, _Minute_, _Sekunde_, _HunderstelSekunde_ * 10);
    }

    //~ SET METHODEN ==========================================================

    /**
     * Setzt den Zeitstempel für das Einzeltelegramm aller Funktionsgruppen. Der Wert ist die Zeitdifferenz zum
     * 01.01.1970 (Mitternacht) in Millisekunden in der Zeitzone UTC.
     * @param timeStamp Zeitdifferenz zum 01.01.1970 (Mitternacht) in Millisekunden in der Zeitzone UTC.
     */
    protected void setTimeStamp(long timeStamp) {
        _timeStamp = timeStamp;
    }

    /**
     * Ermittelt aus den übergebenen Zeitangaben eines TLS-Zeitstempel-DE-Blocks einen gültigen Zeitstempel und stellt
     * das Ergebnis über die Methode {@link #getTimeStamp} zur Verfügung.
     *
     * @param tlsYear        Jahr aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem
     *                       Rechnerdatum ermittelt wird. Angabe kann sowohl vierstellig (z. B. 2004), als auch ohne
     *                       Jahrhundert erfolgen (z.B. 4). Im letzeren Fall wird der Wert automatisch auf 2004
     *                       angepasst (tlsYear  + 2000, wenn 0 kleiner gleich  tlsYear kleiner gleich 100, sonst tlsYear).
     * @param tlsMonth       Monat aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem
     *                       Rechnerdatum ermittelt wird.
     * @param tlsDay         Tag aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem Rechnerdatum
     *                       ermittelt wird.
     * @param tlsHourAndDST  Stundenwert mit Sommerzeitbit aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert
     *                       aus aktuellem Rechnerdatum ermittelt wird. Der übergebene Wert kann sowohl zu einer
     *                       Anpassung aufgrund der Sommerzeit als auch bei Tageswechseln führen. Läuft der Rechner in
     *                       Sommerzeit, die Streckensation aber nicht, wird der übergebene Wert um ein Stunde nach oben
     *                       korrigiert. Läuft der Rechner auf Winterzeit aber die Streckenstation in Sommerzeit, wird
     *                       entsprechend der übergebene Wert um eine Stunde nach unten korrigiert. Zudem wird der Tag
     *                       korrigiert, wenn Telegramm vom Vortag erkannt wird, bzw. Streckenstation schon den
     *                       Tageswechsel durchgeführt hat, der Rechner aber noch nicht. Die Korrektur des Tages erfolgt
     *                       nur, wenn Jahr, Monat und Tag nicht vorgegeben wurden (alle drei Werte also auf -1).
     * @param tlsMinute      Minute aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem
     *                       Rechnerdatum ermittelt wird.
     * @param tlsSecond      Sekunde aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem
     *                       Rechnerdatum ermittelt wird.
     * @param tlsMillisecond Millisekunde aus TLS-Telegramm oder -1. Angabe von -1 bedeutet, dass Wert aus aktuellem
     *                       Rechnerdatum ermittelt wird.
     */
    protected void setTimeStamp(int tlsYear, int tlsMonth, int tlsDay, int tlsHourAndDST, int tlsMinute, int tlsSecond, int tlsMillisecond) {
        GregorianCalendar gregorianCalendar = new GregorianCalendar();
        long              oldTimeStamp      = getTimeStamp();

        gregorianCalendar.setTimeInMillis(oldTimeStamp);

        boolean isTlsDST = false;
        int     tlsHour  = -1;

        if (tlsHourAndDST >= 0) {
            isTlsDST = (tlsHourAndDST & 0x80) > 0;  // Sommerzeitbit TLS gesetzt (DaylightSummerTime)
            tlsHour  = (tlsHourAndDST & 0x7F);      // Stundenanteil ermitteln aus TLS-Stundenwert (DST ausblenden)
        }

        boolean isDST       = (gregorianCalendar.get(Calendar.DST_OFFSET) > 0);  // Sommerzeit (DaylightSummerTime)
        int     year        = gregorianCalendar.get(Calendar.YEAR);
        int     month       = gregorianCalendar.get(Calendar.MONTH);
        int     day         = gregorianCalendar.get(Calendar.DAY_OF_MONTH);
        int     hour        = gregorianCalendar.get(Calendar.HOUR_OF_DAY);
        int     minute      = gregorianCalendar.get(Calendar.MINUTE);
        int     second      = gregorianCalendar.get(Calendar.SECOND);
        int     millisecond = gregorianCalendar.get(Calendar.MILLISECOND);

        // Ergänzung Vorschlag Drapp/Swarco, Fehler #0000087 im NERZ-Bugtracker
        if (isTlsDST)  // wenn ein Wert gesetzt wird, dann muss bei Sommerzeit DST_OFFSET nochmals gesetzt werden
        {
            gregorianCalendar.set(Calendar.DST_OFFSET, 3600000);
        }

        if (tlsYear >= 0) {
            if (tlsYear < 100) {
                year = tlsYear + 2000;  // Fehlt das Jahrtausend/Jahrhundert wird 2000 addiert.
            } else {
                year = tlsYear;
            }

            gregorianCalendar.set(Calendar.YEAR, year);
        }

        if (tlsMonth >= 0) {
            month = tlsMonth - 1;  // Hier wird 1 abgezogen, weil die Monate von 0 bis 11 numeriert werden.
            gregorianCalendar.set(Calendar.MONTH, month);
        }

        if (tlsDay >= 0) {
            day = tlsDay;
            gregorianCalendar.set(Calendar.DAY_OF_MONTH, day);
        }

        if (tlsHour >= 0) {
            gregorianCalendar.set(Calendar.HOUR_OF_DAY, tlsHour);

            // Sind Jahr, Monat und Tag nicht im Telegrammzeitstempel enthalten, dann Vor-/Nachlauf
            // Rechner-Zeit/SST-Zeit beim Tageswechsel berücksichtigen ==> Anpassung des Tages
            if ((tlsDay < 0)
                && (tlsMonth < 0)
                && (tlsYear < 0)) {
                if ((hour == 23)
                    && (tlsHour == 0)) {          // SST ist schon über den Tageswechsel
                    gregorianCalendar.add(Calendar.DAY_OF_MONTH, 1);
                } else if ((hour == 0)
                           && (tlsHour == 23)) {  // SST ist noch auf altem Tag
                    gregorianCalendar.add(Calendar.DAY_OF_MONTH, -1);
                }
            }

            // Laut BitCtrl wird bereits beim Setzen des Datums aus dem TLS-Telegramm über den
            // greorgianCalender das isDST entsprechend angepaßt, also die Stundenkorrektur bereits durchgeführt.
            // Andererseits wurde diese Abfrage in MW von Logos eigefügt, damit die Sommerzeitanpassung funktioniert.
            // Gelobt sei die Einführung der Sommerzeit....
            // Sommer-/Winterzeitverschiebung Rechner gegenüber SST berücksichtigen...
            // 30.11.2011 Korrekturabfrage entfernt, führt u.a. in NI zu Fehler
            //
            // Ab Version 5.4.7 kann die Sonderbehandlung mittels Aufrufparameter eingeschaltet werden:
            //
            // -sonderBehandlungZeitStempelFuerUZinDSTOhneDSTFlag=ja [Default ist nein]
            //
            // siehe auch Fehlereintrag #207 im NERZ-Bugtracker
            // (https://www.nerz-ev.de/nerz-fehlerverwaltung/view_all_set.php?sort=project_id&dir=DESC&type=2)
            //
            if (TlsApplication.isSonderBehandlungZeitStempelFuerUnterschiedeDSTbeiUZundSST()) {
                if ((isDST)
                    && (!isTlsDST)) {             // Rechner Sommerzeit, SST Winterzeit ==> Verschiebung + 1 Stunde
                    gregorianCalendar.add(Calendar.HOUR_OF_DAY, 1);
                } else if ((!isDST)
                           && (isTlsDST)) {       // Rechner Winterzeit, SST Sommerzeit ==> Verschiebung - 1 Stunde
                    gregorianCalendar.add(Calendar.HOUR_OF_DAY, -1);
                }
            }
        }

        if (tlsMinute >= 0) {
            minute = tlsMinute;
            gregorianCalendar.set(Calendar.MINUTE, minute);
        }

        if (tlsSecond >= 0) {
            second = tlsSecond;
            gregorianCalendar.set(Calendar.SECOND, second);
        }

        if (tlsMillisecond >= 0) {
            millisecond = tlsMillisecond;
            gregorianCalendar.set(Calendar.MILLISECOND, millisecond);
        }

        setTimeStamp(gregorianCalendar.getTimeInMillis());

        // Debugausgaben
        String futureTimeStamp = "Zeitstempel ok.";

        if ((oldTimeStamp - getTimeStamp()) < 0) {
            futureTimeStamp = "Zeit liegt in der Zukunft: " + (getTimeStamp() - oldTimeStamp) / 1000 + " Sekunden.\n";
        }

        debug.finest("--------------------------------------------------------------------------------------------------" + "\n" + "Aktuelle Zeit            : " + new Date(oldTimeStamp).toString() + " (Sommerzeit: " + isDST + ")\n" + "TLS-Zeitstempel          : " + "{" + "(Jahr: " + tlsYear + ")," + "(Monat: " + tlsMonth + ")," + "(Tag: " + tlsDay + ")," + "(Stunde: " + tlsHour + ")," + "(Sommerzeit: " + isTlsDST + ")," + "(Minute: " + tlsMinute + ")," + "(Sekunde: " + tlsSecond + ")," +
                     "(MilliSekunde: " + tlsMillisecond + ")}" + "\n" + "Setze Millisekunden      : " + getTimeStamp() + "\n" + "Setze Zeitstempel auf    : " + new Date(getTimeStamp()).toString() + "\n" + futureTimeStamp);
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 25.06.16
